'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var substance = require('substance');

const START = 'START';
const END = 'END';
const EPSILON = 'EPSILON';
const TEXT = 'TEXT';

class DFA {
  constructor (transitions) {
    if (!transitions || Object.keys(transitions).length === 0) {
      transitions = { START: { EPSILON: END } };
    }
    this.transitions = transitions;
  }

  consume (state, id) {
    const T = this.transitions;
    // e.g. this happens, if the state is already END
    // and more tokens are coming
    if (!T[state]) return -1
    let nextState = T[state][id];
    if (nextState !== undefined) {
      return nextState
    }
    while (T[state][EPSILON] !== undefined) {
      state = T[state][EPSILON];
      if (state === END) {
        return -1
      }
      nextState = T[state][id];
      if (nextState !== undefined) {
        return nextState
      }
    }
    return -1
  }

  canConsume (state, id) {
    let nextState = this.consume(state, id);
    return (nextState !== -1)
  }

  isFinished (state) {
    const T = this.transitions;
    if (state === 'END') return true
    // if the state is invalid
    if (!T[state]) return false
    while (T[state][EPSILON] !== undefined) {
      state = T[state][EPSILON];
      if (state === 'END') return true
    }
    return false
  }

  // Helpers to analyze

  // generates all sets of tokens, reached on all different paths
  _tokensByPath () {
    const result = [];
    const transitions = this.transitions;
    if (!transitions) return []

    // group start edges by follow state
    let first = {};
    substance.forEach(transitions[START], (to, token) => {
      if (!first[to]) first[to] = [];
      first[to].push(token);
    });

    let visited = { START: true, END: true };
    substance.forEach(first, (tokens, state) => {
      // walk all states that can be reached on this path
      // and collect all tokens
      // we consider them as potential siblings, as they
      // can co-occur at the same level
      let _siblings = {};
      tokens.forEach((t) => {
        if (t !== EPSILON) {
          _siblings[t] = true;
        }
      });
      let stack = [state];
      while (stack.length > 0) {
        let from = stack.pop();
        if (state === END) continue
        visited[from] = true;
        let T = transitions[from];
        if (!T) throw new Error(`Internal Error: no transition from state ${from}`)
        let tokens = Object.keys(T);
        for (let i = 0; i < tokens.length; i++) {
          const token = tokens[i];
          const to = T[token];
          if (!visited[to]) stack.push(to);
          if (token !== EPSILON) {
            _siblings[token] = true;
          }
        }
      }
      let _siblingTokens = Object.keys(_siblings);
      if (_siblingTokens.length > 0) {
        result.push(_siblingTokens);
      }
    });
    return result
  }
}

DFA.START = START;
DFA.END = END;
DFA.EPSILON = EPSILON;
DFA.TEXT = TEXT;

const START$1 = DFA.START;
const END$1 = DFA.END;
const EPSILON$1 = DFA.EPSILON;

/*
  DFABuilder is essentially a graph implementation
  helping to build DFAs incrementally, by composing smaller
  sub-DFAs.
*/
class DFABuilder {
  constructor (transitions) {
    this.transitions = transitions;
  }

  addTransition (from, to, tokens) {
    if (!this.transitions) this.transitions = {};
    if (!substance.isArray(tokens)) tokens = [tokens];
    tokens.forEach(token => _addTransition(this.transitions, from, to, token));
    return this
  }

  /*
    Creates a new DFA with all state values shifted by a given offset.
    Used when appending this DFA to another one (-> sequence)

    ```
    Expression: A,B

    Graph:  S - [A] -> E
            S - [B] -> E
            =
            S - [A] -> N - [B] -> E
    ```
  */
  append (other) {
    if (this.transitions && other.transitions) {
      let t1 = substance.cloneDeep(this.transitions);
      let t2 = substance.cloneDeep(other.transitions);
      // we need to be careful with EPSILON transitions
      // so that don't end up having a NDFA.
      let firstIsOptional = Boolean(t1[START$1][EPSILON$1]);
      let secondIsOptional = Boolean(t2[START$1][EPSILON$1]);

      if (firstIsOptional) {
        // we remove the epsilon transition from the first
        // as it would be in the way for the transformations done below
        delete t1[START$1][EPSILON$1];
      }
      // for the concatenation we insert a new state and adapt
      // the transitions of the first and second DFA to use the new state
      let newState = substance.uuid();
      // let transitions of the first going to END
      // now point to the new state
      substance.forEach(t1, (T) => {
        substance.forEach(T, (to, token) => {
          if (to === END$1) {
            T[token] = newState;
          }
        });
      });
      // If the first is optional we add transitions from
      // START to the new state
      if (firstIsOptional) {
        substance.forEach(t2[START$1], (to, token) => {
          _addTransition(t1, START$1, to, token);
        });
      }
      // for concatenation we let transitions of the second DFA
      // going from and to START now go from and to the new state
      t2[newState] = t2[START$1];
      substance.forEach(t2, (T) => {
        substance.forEach(T, (to, token) => {
          if (to === START$1) {
            T[token] = newState;
          }
        });
      });
      delete t2[START$1];
      // and now we can merge in the transitions
      substance.forEach(t2, (T, from) => {
        substance.forEach(T, (to, token) => {
          _addTransition(t1, from, to, token);
        });
      });
      // finally we add back an EPSILON transition
      // if both DFAs were optional
      if (firstIsOptional && secondIsOptional) {
        _addTransition(t1, START$1, END$1, EPSILON$1);
      }
      this.transitions = t1;
    } else if (other.transitions) {
      this.transitions = substance.cloneDeep(other.transitions);
    }
    return this
  }

  /*
    Merges to DFAs.

    Used to implement choices.

    ```
    Expression: A | B

    Graph:     - [A] -
              /       \
            S          > E
              \       /
               - [B] -

    ```
  */
  merge (other) {
    if (this.transitions && other.transitions) {
      let t1 = this.transitions;
      let t2 = other.transitions;
      substance.forEach(t2, (T, from) => {
        substance.forEach(T, (to, token) => {
          _addTransition(t1, from, to, token);
        });
      });
    } else if (other.transitions) {
      this.transitions = substance.cloneDeep(other.transitions);
    }
    return this
  }

  /*
    Creates a new DFA with same transitions
    plus an EPSILON transition from start to END

    ```
    Expression: A?

    Graph:     - [A] -
              /       \
            S          > E
              \       /
               -  ε  -
    ```
  */
  optional () {
    let dfa = new DFABuilder(substance.cloneDeep(this.transitions));
    if (this.transitions) {
      dfa.addTransition(START$1, END$1, EPSILON$1);
    }
    return dfa
  }

  /*
    Creates a new DFA representing (A)*

    ```
    Expression: A* = (A+)?

                       /-[A]-\
                       |     |
                        \   /
                         v /
    Graph:   S -- [A] --> 1  -- ε -->  E
              \                    /
               \   --    ε   --   /

    ```
  */
  kleene () {
    let dfa = this.plus();
    return dfa.optional()
  }

  /*
    Creates a new DFA representing (...)+ by concatenating this
    with a kleene version: A+ = A A*

    ```
    Expression: (A)+ (sequence and reflexive edge)

                       /-[A]-\
                       |     |
                        \   /
                         v /
    Graph:  S -- [A] -->  N  -- ε -> E

    ```
  */
  plus () {
    let dfa;
    if (this.transitions) {
      let t1 = substance.cloneDeep(this.transitions);
      // there might exist an EPSILON transition already
      // which we must remove to fulfill our internal
      // assumption that there is only one EPSILON transition from
      // START going to END
      const isOptional = Boolean(t1[START$1][EPSILON$1]);
      delete t1[START$1][EPSILON$1];
      // introduce a new state
      // and let all 'ending' edges point to new state
      let newState = substance.uuid();
      substance.forEach(t1, (T) => {
        substance.forEach(T, (to, token) => {
          if (to === END$1) {
            T[token] = newState;
          }
        });
      });
      // add 'ending' EPSILON transition
      _addTransition(t1, newState, END$1, EPSILON$1);
      // copy all starting edges
      substance.forEach(t1[START$1], (to, token) => {
        _addTransition(t1, newState, to, token);
      });
      // recover 'optional'
      if (isOptional) {
        _addTransition(t1, START$1, END$1, EPSILON$1);
      }
      dfa = new DFABuilder(t1);
    } else {
      dfa = new DFABuilder(substance.cloneDeep(this.transitions));
    }
    return dfa
  }
}

DFABuilder.singleToken = function (token) {
  let dfa = new DFABuilder();
  dfa.addTransition(START$1, END$1, token);
  return dfa
};

function _addTransition (transitions, from, to, token) {
  let T = transitions[from];
  if (!T) {
    transitions[from] = T = {};
  }
  if (token === EPSILON$1 && from === START$1 && to !== END$1) {
    throw new Error('The only EPSILON transition from START must be START->END')
  }
  if (T[token] && T[token] !== to) {
    console.error('Token %s already used. Ignoring this transition.', token);
    return
    // throw new Error('Token already used in this state')
  }
  T[token] = to;
}

const { START: START$2, END: END$2, TEXT: TEXT$1, EPSILON: EPSILON$2 } = DFA;

// retains the structured representation
// and compiles a DFA for efficient processing
class Expression {
  // TODO: why does the expression need a name?
  constructor (name, root) {
    this.name = name;
    this.root = root;

    this._initialize();
  }

  _initialize () {
    this._compile();
  }

  toString () {
    return this.root.toString()
  }

  isAllowed (tagName) {
    return Boolean(this._allowedChildren[tagName])
  }

  /*
    Some structures get compiled into a DFA, for instance.
  */
  _compile () {
    this.root._compile();
  }

  _describeError (state, token) {
    let msg = [];
    if (token !== TEXT$1) {
      if (!this.isAllowed(token)) {
        msg.push(`<${token}> is not valid in <${this.name}>\nSchema: ${this.toString()}`);
      } else {
        // otherwise just the position is wrong
        msg.push(`<${token}> is not allowed at the current position in <${this.name}>.\n${this.toString()}`);
      }
    } else {
      msg.push(`TEXT is not allowed at the current position: ${state.trace.join(',')}\n${this.toString()}`);
    }
    return msg.join('')
  }
}

function createExpression (name, root) {
  if (root instanceof Interleave) {
    return new InterleaveExpr(name, root)
  } else {
    return new DFAExpr(name, root)
  }
}

class DFAExpr extends Expression {
  getInitialState () {
    return {
      dfaState: START$2,
      errors: [],
      trace: []
    }
  }

  consume (state, token) {
    const dfa = this.dfa;
    let oldState = state.dfaState;
    let newState = dfa.consume(oldState, token);
    state.dfaState = newState;
    if (newState === -1) {
      state.errors.push({
        msg: this._describeError(state, token),
        // HACK: we want to have the element with the errors
        // but actually, here we do not know about that context
        el: state.el
      });
      return false
    } else {
      state.trace.push(token);
      return true
    }
  }

  isFinished (state) {
    return this.dfa.isFinished(state.dfaState)
  }

  _initialize () {
    super._initialize();

    this._computeAllowedChildren();
  }

  _compile () {
    super._compile();
    this.dfa = new DFA(this.root.dfa.transitions);
  }

  _computeAllowedChildren () {
    this._allowedChildren = _collectAllTokensFromDFA(this.dfa);
  }

  _isValid (_tokens) {
    let state = this.getInitialState();
    for (let i = 0; i < _tokens.length; i++) {
      const token = _tokens[i];
      // Note: there might be some elements which
      // are not relevant, such as empty text nodes
      // or comments etc.
      if (!token) continue
      if (!this.consume(state, token)) {
        return false
      }
    }
    return this.isFinished(state)
  }
}

function _collectAllTokensFromDFA (dfa) {
  // Note: collecting all children
  const children = {};
  if (dfa.transitions) {
    substance.forEach(dfa.transitions, (T) => {
      Object.keys(T).forEach((tagName) => {
        if (tagName === EPSILON$2) return
        children[tagName] = true;
      });
    });
  }
  return children
}

class InterleaveExpr extends Expression {
  getInitialState () {
    const dfas = this.dfas;
    const dfaStates = new Array(dfas.length);
    dfaStates.fill(START$2);
    return {
      dfaStates,
      errors: [],
      trace: [],
      // maintain the index of the dfa which has been consumed the last token
      lastDFA: 0
    }
  }

  consume (state, token) {
    const idx = this._findNextDFA(state, token);
    if (idx < 0) {
      state.errors.push({
        msg: this._describeError(state, token)
      });
      return false
    } else {
      const dfa = this.dfas[idx];
      const oldState = state.dfaStates[idx];
      const newState = dfa.consume(oldState, token);
      state.dfaStates[idx] = newState;
      state.trace.push(token);
      return true
    }
  }

  isFinished (state) {
    const dfas = this.dfas;
    for (let i = 0; i < dfas.length; i++) {
      const dfa = dfas[i];
      const dfaState = state.dfaStates[i];
      if (!dfa.isFinished(dfaState)) {
        return false
      }
    }
    return true
  }

  _initialize () {
    super._initialize();

    this._computeAllowedChildren();
  }

  _compile () {
    super._compile();

    this.blocks = this.root.blocks;
    this.dfas = this.blocks.map(b => new DFA(b.dfa.transitions));
  }

  _computeAllowedChildren () {
    this._allowedChildren = Object.assign(...this.blocks.map((block) => {
      return _collectAllTokensFromDFA(block.dfa)
    }));
  }

  _findNextDFA (state, token) {
    console.assert(state.dfaStates.length === this.dfas.length);
    const dfas = this.dfas;
    for (let i = 0; i < state.dfaStates.length; i++) {
      const dfa = dfas[i];
      const dfaState = state.dfaStates[i];
      if (dfa.canConsume(dfaState, token)) {
        return i
      }
    }
    return -1
  }
}

class Token {
  constructor (name) {
    this.name = name;
  }

  toString () {
    return this.name
  }

  _compile () {
    this.dfa = DFABuilder.singleToken(this.name);
  }
}

class GroupExpression {
  constructor (blocks) {
    this.blocks = blocks;
  }

  toString () {
    return '(' + this.blocks.map(b => b.toString()).join(this.token) + ')'
  }
}

/*
  (a|b|c)
*/
class Choice extends GroupExpression {
  // copy () {
  //   return new Choice(this.blocks.map(b => b.copy()))
  // }

  // _normalize () {
  //   const blocks = this.blocks
  //   for (let i = blocks.length - 1; i >= 0; i--) {
  //     let block = blocks[i]
  //     block._normalize()
  //     // unwrap doubled Choices
  //     if (block instanceof Choice) {
  //       blocks.splice(i, 1, ...(block.blocks))
  //     }
  //   }
  // }

  _compile () {
    let dfa = new DFABuilder();
    this.blocks.forEach((block) => {
      if (block instanceof Token) {
        dfa.addTransition(START$2, END$2, block.name);
      } else if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      } else {
        if (!block.dfa) {
          block._compile();
        }
        dfa.merge(block.dfa);
      }
    });
    this.dfa = dfa;
    return dfa
  }

  get token () { return Choice.token }

  static get token () { return '|' }
}

/*
  (a,b,c) (= ordered)
*/
class Sequence extends GroupExpression {
  // copy () {
  //   return new Sequence(this.blocks.map(b => b.copy()))
  // }

  _compile () {
    let dfa = new DFABuilder();
    this.blocks.forEach((block) => {
      if (block instanceof Token) {
        dfa.append(DFABuilder.singleToken(block.name));
      } else if (block instanceof Interleave) {
        throw new Error('Nested interleave blocks are not supported.')
      } else {
        if (!block.dfa) {
          block._compile();
        }
        dfa.append(block.dfa);
      }
    });
    this.dfa = dfa;
    return dfa
  }

  get token () { return Sequence.token }

  static get token () { return ',' }
}

/*
  ~(a,b,c) (= unordered)
*/
class Interleave extends GroupExpression {
  // copy () {
  //   return new Interleave(this.blocks.map(b => b.copy()))
  // }

  toString () {
    return '(' + this.blocks.map(b => b.toString()).join(', ') + ')[unordered]'
  }

  _normalize () {}

  _compile () {
    this.blocks.forEach(block => block._compile());
  }

  get token () { return Interleave.token }

  static get token () { return '~' }
}

class BlockExpression {
  constructor (block) {
    this.block = block;
  }

  toString () {
    return this.block.toString() + this.token
  }
}

/*
  ()?
*/
class Optional extends BlockExpression {
  // copy () {
  //   return new Optional(this.block.copy())
  // }

  _compile () {
    const block = this.block;
    if (block instanceof Interleave) {
      throw new Error('Nested interleave blocks are not supported.')
    }
    if (!block.dfa) {
      block._compile();
    }
    this.dfa = block.dfa.optional();
    return this.dfa
  }

  get token () { return Optional.token }

  static get token () { return '?' }
}

/*
  ()*
*/
class Kleene extends BlockExpression {
  // copy () {
  //   return new Kleene(this.block.copy())
  // }

  _compile () {
    const block = this.block;
    if (block instanceof Interleave) {
      throw new Error('Nested interleave blocks are not supported.')
    }
    if (!block.dfa) {
      block._compile();
    }
    this.dfa = block.dfa.kleene();
    return this.dfa
  }

  get token () { return Kleene.token }

  static get token () { return '*' }
}

/*
  ()+
*/
class Plus extends BlockExpression {
  // copy () {
  //   return new Plus(this.block.copy())
  // }

  _compile () {
    const block = this.block;
    if (block instanceof Interleave) {
      throw new Error('Nested interleave blocks are not supported.')
    }
    if (!block.dfa) {
      block._compile();
    }
    this.dfa = block.dfa.plus();
    return this.dfa
  }

  get token () { return Plus.token }

  static get token () { return '+' }
}

const { TEXT: TEXT$2 } = DFA;

function analyze (elementSchemas) {
  substance.forEach(elementSchemas, elementSchema => {
    Object.assign(elementSchema, {
      children: {},
      parents: {},
      siblings: {},
      usedInlineBy: {},
      usedStructuredBy: {}
    });
  });
  substance.forEach(elementSchemas, elementSchema => {
    _analyzeElementSchema(elementSchema, elementSchemas);
  });
}

/*
 We use this to detect automatically, if an
 element is used as a text node or an element node,
 or both at the same time.
*/
function _analyzeElementSchema (elementSchema, elementSchemas) {
  const expr = elementSchema.expr;
  const name = elementSchema.name;
  if (!expr) return
  let _siblings = [];
  if (expr instanceof DFAExpr) {
    if (expr.dfa) {
      _siblings = expr.dfa._tokensByPath();
    }
  } else if (expr instanceof InterleaveExpr) {
    expr.dfas.forEach((dfa) => {
      if (dfa) {
        _siblings = _siblings.concat(dfa._tokensByPath());
      }
    });
  }

  let hasText = false;
  let hasElements = false;
  _siblings.forEach((tagNames) => {
    // register each other as parent and children
    let _hasText = tagNames.indexOf(TEXT$2) >= 0;
    let _hasElements = (!_hasText && tagNames.length > 0);
    if (_hasText) {
      hasText = true;
    }
    if (_hasElements) {
      hasElements = true;
    }
    tagNames.forEach((tagName) => {
      const childSchema = elementSchemas[tagName];
      if (!childSchema) return
      childSchema.parents[name] = true;
      elementSchema.children[tagName] = true;
      // Note: we store siblings, grouped by parent
      elementSchema.siblings[name] = tagNames;
      if (_hasElements) childSchema.usedStructuredBy[name] = true;
      if (_hasText) childSchema.usedInlineBy[name] = true;
    });
  });
  // TODO: document what these fields are used for.
  if (hasElements) elementSchema.isStructured = true;
  if (hasText) elementSchema.isText = true;
  if (!elementSchema.type) {
    if (hasText) {
      elementSchema.type = 'text';
    } else {
      elementSchema.type = 'element';
    }
  }
}

class ElementSchema {
  constructor (name, type, attributes, expr) {
    this.name = name;
    this.type = type;
    this.attributes = attributes;
    this.expr = expr;
  }
}

function _isTextNodeEmpty (el) {
  return Boolean(/^\s*$/.exec(el.textContent))
}

function _validateElement (elementSchema, el) {
  let errors = [];
  let valid = true;
  if (!elementSchema) {
    return {
      errors: [ {
        msg: `Unknown tag <${el.tagName}>.`,
        el
      } ],
      ok: false
    }
  }
  // Elements
  if (elementSchema.type === 'external' || elementSchema.type === 'not-implemented') ; else {
    let res = _checkChildren(elementSchema, el);
    if (!res.ok) {
      errors = errors.concat(res.errors);
      valid = false;
    }
  }
  return {
    errors,
    ok: valid
  }
}

function _checkChildren (elementSchema, el) {
  // Don't validate external nodes
  if (elementSchema.type === 'external' || elementSchema.type === 'not-implemented') {
    return true
  }
  const isText = elementSchema.type === 'text';
  const expr = elementSchema.expr;
  const state = expr.getInitialState();
  const iterator = el.getChildNodeIterator();
  let valid = true;
  let tokenCount = 0;
  while (valid && iterator.hasNext()) {
    const childEl = iterator.next();
    let token;
    if (childEl.isTextNode()) {
      // Note: skipping empty text being child node of elements
      if (_isTextNodeEmpty(childEl)) {
        continue
      } else {
        token = DFA.TEXT;
      }
    } else if (childEl.isElementNode()) {
      token = childEl.tagName;
    } else if (childEl.getNodeType() === 'cdata') {
      // CDATA elements are treated as a TEXT fragment
      token = DFA.TEXT;
    } else {
      continue
    }
    tokenCount++;
    if (!expr.consume(state, token)) {
      valid = false;
    }
  }
  // add the element to the errors
  if (state.errors.length > 0) {
    state.errors.forEach((err) => {
      err.el = el;
    });
  }
  const isFinished = expr.isFinished(state);
  if (valid && !isFinished) {
    if (isText && tokenCount === 0) ; else {
      state.errors.push({
        msg: `<${el.tagName}> is incomplete.\nSchema: ${expr.toString()}`,
        el
      });
      valid = false;
    }
  }
  if (valid) {
    state.ok = true;
  }
  return state
}

class XMLSchema {
  constructor (elementSchemas, startElement, publicId, dtd) {
    if (!elementSchemas[startElement]) {
      throw new Error('startElement must be a valid element.')
    }
    this._elementSchemas = {};
    this.startElement = startElement;
    this.publicId = publicId;
    this.dtd = dtd;
    // wrap schemas into ElementSchemas
    substance.forEach(elementSchemas, (spec, name) => {
      this._elementSchemas[name] = new ElementSchema(spec.name, spec.type, spec.attributes, spec.expr);
    });
  }

  getIdAttribute () {
    return 'id'
  }

  getTagNames () {
    return Object.keys(this._elementSchemas)
  }

  getDocTypeParams () {
    return [this.startElement, this.publicId, this.dtd]
  }

  getElementSchema (name) {
    return this._elementSchemas[name]
  }

  getStartElement () {
    return this.startElement
  }

  validateElement (el) {
    let tagName = el.tagName;
    let elementSchema = this.getElementSchema(tagName);
    return _validateElement(elementSchema, el)
  }
}

/**
 * Look up a RNG file from the current directory or a list of search directories.
 *
 * @param {*} fs
 * @param {*} rngFileName
 * @param {*} currentDir
 * @param {*} searchDirs
 */
function _lookupRNG (fs, rngFileName, currentDir, searchDirs) {
  let rngPath;
  // 1. Try if the file can be found directly
  rngPath = rngFileName;
  if (fs.existsSync(rngPath)) {
    return rngPath
  }
  // 2. Try the current directory
  rngPath = currentDir + '/' + rngFileName;
  if (fs.existsSync(rngPath)) {
    return rngPath
  }
  // 3. Try the search directories
  for (let i = 0; i < searchDirs.length; i++) {
    rngPath = searchDirs[i] + '/' + rngFileName;
    if (fs.existsSync(rngPath)) {
      return rngPath
    }
  }
}

function _expandIncludes (fs, path, currentDir, searchDirs, grammarEl) {
  let includes = grammarEl.findAll('include');
  if (includes.length === 0) return false
  includes.forEach(include => {
    const parent = include.parentNode;
    const href = include.attr('href');
    const rngPath = _lookupRNG(fs, href, currentDir, searchDirs);
    if (!rngPath) throw new Error(`Could not find ${href}`)
    const rngStr = fs.readFileSync(rngPath, 'utf8');
    const rng = substance.DefaultDOMElement.parseXML(rngStr, 'full-doc');
    const _grammarEl = rng.find('grammar');
    if (!_grammarEl) throw new Error('No grammar element found')
    let rngDir = path.dirname(rngPath);
    // expand the grammar recursively
    _expandIncludes(fs, path, rngDir, searchDirs, _grammarEl);
    // now replace the include element with the content of the expanded grammar
    _grammarEl.children.forEach((child) => {
      parent.insertBefore(child, include);
    });
    include.remove();
  });
  return true
}

/*
  Loads a RNG with all dependencies into a DOM element
*/
function _loadRNG (fs, path, rngFile, searchDirs) {
  if (!substance.isArray(searchDirs)) searchDirs = [searchDirs];
  let rngDir = path.dirname(rngFile);
  let rngStr = fs.readFileSync(rngFile, 'utf8');
  const rng = substance.DefaultDOMElement.parseXML(rngStr, 'full-doc');
  const grammarEl = rng.find('grammar');
  _expandIncludes(fs, path, rngDir, searchDirs, grammarEl);
  return rng
}

// import prettyPrintXML from './prettyPrintXML'

const TEXT$3 = DFA.TEXT;

/*
  We use regular RNG, with slight restrictions plus custom extensions,
  and compile it into our internal format.
*/
function _compileRNG (fs, path, rngFile, searchDirs) {
  let rng = _loadRNG(fs, path, rngFile, searchDirs);
  let grammar = rng.find('grammar');
  if (!grammar) throw new Error('<grammar> not found.')
  // collect all definitions, allowing for custom overrides
  _registerDefinitions(grammar);
  // turn the RNG schema into our internal data structure
  let transformedGrammar = _transformRNG(grammar);
  // console.log(prettyPrintXML(transformedGrammar))
  let xmlSchema = _compile(transformedGrammar);
  return xmlSchema
}

/* Registration of <define> elements */

function _registerDefinitions (grammar) {
  let defs = {};
  // NOTE: definitions are only considered on the top level
  grammar.children.forEach(child => {
    const tagName = substance.nameWithoutNS(child.tagName);
    if (tagName === 'define') {
      _processDefine(child, defs);
    }
  });
  grammar.defs = defs;
}

function _processDefine (el, defs) {
  const name = el.attr('name');
  const combine = el.attr('combine');
  if (combine === 'interleave') {
    if (defs[name]) {
      defs[name].append(el.children);
    } else {
      defs[name] = el;
    }
  } else {
    if (defs[name]) ;
    defs[name] = el;
  }
}

/* Transformation of RNG into internal representation */
function _transformRNG (grammar) {
  const $$ = grammar.createElement.bind(grammar);
  // remove everything elements from the grammar that have been tagged as 's:removed'
  grammar.findAll('removed').forEach(el => {
    let name = el.attr('name');
    grammar.findAll(`element[name="${name}"]`).forEach(el => {
      // console.log('removing <element>', name)
      el.remove();
    });
    grammar.findAll(`ref[name="${name}"]`).forEach(el => {
      // console.log('removing <ref>', name)
      el.remove();
    });
  });

  const elements = {};
  const defs = grammar.defs;
  const elementDefinitions = grammar.findAll('define > element');
  const doc = substance.DefaultDOMElement.createDocument('xml');
  const newGrammar = doc.createElement('grammar');

  // record all not implemented ones
  // we will allow to use them, but skip their content definition
  const notImplemented = grammar.findAll('not-implemented').reduce((s, el) => {
    let name = el.attr('name');
    if (name) s.add(name);
    return s
  }, new Set());

  // expand definitions
  elementDefinitions.forEach(el => {
    const name = el.attr('name');
    if (!name) throw new Error("'name' is mandatory.")
    let transformed;
    if (notImplemented.has(name)) {
      transformed = $$('element').attr('name', name).attr('type', 'not-implemented');
    } else {
      transformed = _transformElementDefinition(doc, name, el, defs);
    }
    elements[name] = transformed;
    newGrammar.appendChild(transformed);
  });

  // infer element types
  // TODO: do we need this anymore?
  const elementTypes = grammar.findAll('elementType');
  elementTypes.forEach(typeEl => {
    const name = typeEl.attr('name');
    let type = typeEl.attr('s:type') || typeEl.attr('type');
    if (!name || !type) throw new Error('Attributes name and type are mandatory.')
    const element = elements[name];
    if (!element) throw new Error(`Unknown element ${name}.`)
    element.attr('type', type);
  });

  // start element
  const startElement = _extractStart(grammar);
  if (!startElement) throw new Error('<start> is mandatory.')
  newGrammar.appendChild(doc.createElement('start').attr('name', startElement));

  return newGrammar
}

function _transformElementDefinition (doc, name, orig, defs) {
  let el = doc.createElement('element').attr('name', name);
  // TODO: try to separate attributes from children
  // now go through all children and wrap them into attributes and children
  let attributes = doc.createElement('attributes');
  let children = doc.createElement('children');
  orig.children.forEach((child) => {
    let block = _transformBlock(doc, child, defs, {});
    block.forEach((el) => {
      if (el.find('attribute') || el.is('attribute')) {
        attributes.appendChild(el);
      } else {
        children.appendChild(el);
      }
    });
  });
  el.appendChild(attributes);
  el.appendChild(children);

  /*
    Pruning (this is probably very slow!)
    - choice > choice
    - choice with one element
  */
  while (true) {
    // Unwrap nested choices
    let nestedChoice = children.find('choice > choice');
    if (nestedChoice) {
      // unwrap onto parent level
      let parentChoice = nestedChoice.parentNode;
      // TODO: we could use DOM helpers as we do in Texture converters
      let children = nestedChoice.children;
      children.forEach((child) => {
        parentChoice.insertBefore(child, nestedChoice);
      });
      parentChoice.removeChild(nestedChoice);
      continue
    }
    break
  }

  // Simplify singular choices
  let choices = children.findAll('choice');
  for (let i = 0; i < choices.length; i++) {
    let choice = choices[i];
    let children = choice.children;
    if (children.length === 1) {
      choice.parentNode.replaceChild(choice, children[0]);
    }
  }

  let optionalTextEls = children.findAll('optional > text, zeroOrMore > text');
  for (let i = 0; i < optionalTextEls.length; i++) {
    let textEl = optionalTextEls[i];
    let optionalEl = textEl.parentNode;
    if (optionalEl.getChildCount() === 1) {
      optionalEl.parentNode.replaceChild(optionalEl, textEl);
    }
  }

  // remove empty groups
  let groupEls = children.findAll('optional, zeroOrMore, oneOrMore');
  for (let i = 0; i < groupEls.length; i++) {
    let groupEl = groupEls[i];
    if (groupEl.getChildCount() === 0) {
      groupEl.remove();
    }
  }

  return el
}

function _transformBlock (doc, block, defs, visiting = {}) {
  // if a block is a <ref> return the expanded children
  // otherwise clone the block and descend recursively
  const tagName = block.tagName;
  switch (tagName) {
    case 'element': {
      return [doc.createElement('element').attr('name', block.attr('name'))]
    }
    case 'ref': {
      return _expandRef(doc, block, defs, visiting)
    }
    case 'empty':
    case 'notAllowed': {
      return []
    }
    default: {
      // TODO: while this is a valid approach, it could be more efficient
      // to 'reuse' already processed elements (i.e. reuse their DFA)
      // For that reason, I have commented out all occurrences where I used to resuse the DFA
      // being dead code at the moment
      let clone = block.clone(false);
      block.children.forEach((child) => {
        clone.append(_transformBlock(doc, child, defs, visiting));
      });
      return [clone]
    }
  }
}

function _expandRef (doc, ref, defs, visiting = {}) {
  const name = ref.attr('name');
  // Acquire semaphore against cyclic refs
  if (visiting[name]) {
    throw new Error('Cyclic references are not supported.')
  }
  visiting[name] = true;

  const def = defs[name];
  if (!def) throw new Error(`Unknown definition ${name}`)

  let expanded = [];
  let children = def.children;
  children.forEach((child) => {
    let transformed = _transformBlock(doc, child, defs, visiting);
    expanded = expanded.concat(transformed);
  });

  // Releasing semaphore against cyclic refs
  delete visiting[name];
  return expanded
}

function _extractStart (grammar) {
  // for now this is hard wired to work with the start
  // element as defined in JATS 1.1
  const start = grammar.find('start');
  if (!start) {
    throw new Error('<grammar> must have a <start> element')
  }
  // HACK: we assume that there is exactly one ref to
  // an element definition
  const startRef = start.find('ref');
  if (!startRef) {
    throw new Error('Expecting one <ref> inside of <start>.')
  }
  const name = startRef.attr('name');
  return name
}

function _compile (grammar) {
  const schemas = {};
  const elements = grammar.children.filter(el => el.tagName === 'element');
  elements.forEach(element => {
    const name = element.attr('name');
    const attributes = _collectAttributes(element.find('attributes'));
    const children = element.find('children');
    const type = element.attr('type');
    let block = _processChildren(children, grammar);
    let expr = createExpression(name, block);
    let schema = { name, type, attributes, expr };
    schemas[name] = schema;
  });

  // this adds some reflection info and derives the type
  analyze(schemas);

  const start = grammar.find('start');
  if (!start) {
    throw new Error('<start> is mandatory')
  }
  const startElement = start.attr('name');
  if (!startElement) {
    throw new Error('<start> must have "name" set')
  }
  return new XMLSchema(schemas, startElement)
}

function _processChildren (el, grammar) {
  if (!el) return new Sequence([])
  let blocks = _processBlocks(el.children, grammar);
  if (blocks.length === 1) {
    return blocks[0]
  } else {
    return new Sequence(blocks)
  }
}

function _processBlocks (children, grammar) {
  const blocks = [];
  for (var i = 0; i < children.length; i++) {
    const child = children[i];
    // const name = child.attr('name')
    switch (child.tagName) {
      // skip these
      case 'attribute':
      case 'empty':
      case 'notAllowed': {
        break
      }
      case 'element': {
        const elName = child.attr('name');
        blocks.push(new Token(elName));
        break
      }
      case 'text': {
        blocks.push(new Token(TEXT$3));
        break
      }
      case 'ref': {
        const block = _processReference(child, grammar);
        blocks.push(block);
        break
      }
      case 'group': {
        blocks.push(_processSequence(child, grammar));
        break
      }
      case 'choice': {
        const block = _processChoice(child, grammar);
        blocks.push(block);
        break
      }
      case 'optional': {
        const block = new Optional(_processChildren(child, grammar));
        blocks.push(block);
        break
      }
      case 'oneOrMore': {
        const block = new Plus(_processChildren(child, grammar));
        blocks.push(block);
        break
      }
      case 'zeroOrMore': {
        const block = new Kleene(_processChildren(child, grammar));
        blocks.push(block);
        break
      }
      case 'interleave': {
        const block = new Interleave(_processBlocks(child.children, grammar));
        blocks.push(block);
        break
      }
      default:
        throw new Error('Not supported yet: ' + child.tagName)
    }
  }
  return blocks
}

function _processSequence (el, grammar) {
  // TODO: seems that this optimization is not needed any more as references get inlined.
  // looking at _expandRef() it looks as though the corresponding DOMElement gets cloned on recursion
  // see above
  // if (el.expr) return el.expr.copy()
  const blocks = _processBlocks(el.children, grammar);
  el.expr = new Sequence(blocks);
  return el.expr
}

function _processChoice (el, grammar) {
  // if (el.expr) return el.expr.copy()
  let blocks = _processBlocks(el.children, grammar);
  el.expr = new Choice(blocks);
  return el.expr
}

function _processReference (ref, grammar) {
  const name = ref.attr('name');
  const def = grammar.defs[name];
  if (!def) throw new Error(`Illegal ref: ${name} is not defined.`)
  // if (def.expr) return def.expr.copy()
  // Guard for cyclic references
  // TODO: what to do with cyclic refs?
  if (grammar._visiting[name]) {
    throw new Error('Cyclic references are not supported yet')
  }
  grammar._visiting[name] = true;
  const block = _processChildren(def, grammar);
  def.expr = block;
  delete grammar._visiting[name];
  return def.expr
}

function _collectAttributes (el, grammar, attributes = {}) {
  if (!el) return {}
  // ATTENTION: RNG supports more than we do here
  // We just collect all attributes, infering no rules
  let children = el.children;
  for (let i = 0; i < children.length; i++) {
    const child = children[i];
    switch (child.tagName) {
      case 'attribute': {
        const attr = _transformAttribute(child);
        attributes[attr.name] = attr;
        break
      }
      case 'group':
      case 'choice':
      case 'optional':
      case 'oneOrMore':
      case 'zeroOrMore': {
        _collectAttributes(child, grammar, attributes);
        break
      }
      default:
        //
    }
  }
  return attributes
}

function _transformAttribute (el) {
  const name = el.attr('name');
  // TODO: extract all the attribute specs
  return {
    name
  }
}

function validateXML (xmlSchema, dom) {
  let root = dom.find(xmlSchema.getStartElement());
  if (!root) {
    return {
      errors: [{
        msg: 'Start element is missing.',
        el: dom
      }]
    }
  } else {
    return validateElement(xmlSchema, root)
  }
}

function validateElement (xmlSchema, el) {
  let errors = [];
  let valid = true;
  let q = [el];
  while (q.length > 0) {
    let next = q.shift();
    let res = xmlSchema.validateElement(next);
    if (!res.ok) {
      errors = errors.concat(res.errors);
      valid = false;
    }
    if (next.isElementNode()) {
      q = q.concat(next.getChildren());
    }
  }
  return {
    errors: errors,
    ok: valid
  }
}

/**
 * This implementation is creating a minified representation of a Schema.
 *
 * ```
 * [ [<string-literals...>], <rootElId>, [<elements...>] ]
 *
 * element: [ <nameId>, [<attributes...>], <content> ]
 * attribute: <nameId> (temporarily)
 * content: [<type>:'*?+si', []|<nameId> ]
 * ```
 */

class LiteralRegistry {
  constructor () {
    this._map = new Map();
  }

  register (literal) {
    if (!this._map.has(literal)) {
      this._map.set(literal, { literal, count: 0 });
    }
    this._map.get(literal).count++;
  }

  computeRanks () {
    let entries = Array.from(this._map.values());
    entries.sort((a, b) => {
      return b.count - a.count
    });
    let L = entries.length;
    for (let idx = 0; idx < L; idx++) {
      entries[idx].rank = idx;
    }
    this._sortedLiterals = entries.map(e => e.literal);
  }

  getRank (literal) {
    return this._map.get(literal).rank
  }

  getSortedLiterals () {
    return this._sortedLiterals
  }
}

function serializeXMLSchema (xmlSchema) {
  let literalRegistry = new LiteralRegistry();

  function _registerLiterals (o) {
    switch (o.constructor) {
      case XMLSchema: {
        literalRegistry.register(o.startElement);
        o.getTagNames().forEach(name => {
          _registerLiterals(o.getElementSchema(name));
        });
        break
      }
      case ElementSchema: {
        literalRegistry.register(o.name);
        Object.keys(o.attributes).forEach(attrName => {
          // TODO: later we should also register attribute values
          literalRegistry.register(attrName);
        });
        _registerLiterals(o.expr);
        break
      }
      case DFAExpr:
      case InterleaveExpr: {
        _registerLiterals(o.root);
        break
      }
      case Token: {
        literalRegistry.register(o.name);
        break
      }
      case Choice:
      case Sequence:
      case Interleave: {
        o.blocks.forEach(_registerLiterals);
        break
      }
      case Optional:
      case Kleene:
      case Plus: {
        _registerLiterals(o.block);
        break
      }
      default:
        throw new Error('FIXME')
    }
  }
  _registerLiterals(xmlSchema);

  literalRegistry.computeRanks();

  function _encode (o) {
    switch (o.constructor) {
      case XMLSchema: {
        return [
          literalRegistry.getRank(o.startElement),
          o.getTagNames().map(name => {
            return _encode(o.getElementSchema(name))
          })
        ]
      }
      case ElementSchema: {
        return [
          literalRegistry.getRank(o.name),
          o.type === 'text' ? 't' : 'e',
          Object.keys(o.attributes).map(attrName => {
            // TODO: later we should also register attribute values
            return literalRegistry.getRank(attrName)
          }),
          _encode(o.expr)
        ]
      }
      case DFAExpr:
      case InterleaveExpr: {
        return _encode(o.root)
      }
      case Token: {
        return literalRegistry.getRank(o.name)
      }
      case Choice:
      case Sequence:
      case Interleave: {
        return [
          o.token,
          o.blocks.map(_encode)
        ]
      }
      case Optional:
      case Kleene:
      case Plus: {
        return [
          o.token,
          _encode(o.block)
        ]
      }
    }
  }
  let data = { literals: literalRegistry.getSortedLiterals(), schema: _encode(xmlSchema) };

  return JSON.stringify(data)
}

function deserializeXMLSchema (xmlSchemaInput, publicId = '', dtd = '') {
  let data;
  if (substance.isString(xmlSchemaInput)) {
    data = JSON.parse(xmlSchemaInput);
  } else {
    data = xmlSchemaInput;
  }
  let literals = data.literals;
  let schemaData = data.schema;

  function _decodeLiteral (d) {
    return literals[d]
  }

  let startElement = _decodeLiteral(schemaData[0]);
  let elementSchemas = {};

  function _decodeExpression (d) {
    if (substance.isNumber(d)) {
      return new Token(_decodeLiteral(d))
    } else if (substance.isArray(d)) {
      let type = d[0];
      let content = d[1];
      switch (type) {
        case Sequence.token:
          return new Sequence(content.map(_decodeExpression))
        case Interleave.token:
          return new Interleave(content.map(_decodeExpression))
        case Choice.token:
          return new Choice(content.map(_decodeExpression))
        case Optional.token:
          return new Optional(_decodeExpression(content))
        case Plus.token:
          return new Plus(_decodeExpression(content))
        case Kleene.token:
          return new Kleene(_decodeExpression(content))
      }
    } else {
      throw new Error('invalid data')
    }
  }

  function _decodeElementSchemaData (d) {
    let name = _decodeLiteral(d[0]);
    let type = d[1] === 't' ? 'text' : 'element';
    // TODO: at some point we gonna have more complex attribute specs
    let attributes = {};
    d[2].forEach(rank => {
      let literal = _decodeLiteral(rank);
      attributes[literal] = literal;
    });
    let expr = createExpression(name, _decodeExpression(d[3]));
    return new ElementSchema(name, type, attributes, expr)
  }

  schemaData[1].forEach(elementSchemaData => {
    let elementSchema = _decodeElementSchemaData(elementSchemaData);
    elementSchemas[elementSchema.name] = elementSchema;
  });

  let schema = new XMLSchema(elementSchemas, startElement, publicId, dtd);
  return schema
}

exports.DFA = DFA;
exports.DFABuilder = DFABuilder;
exports.XMLSchema = XMLSchema;
exports._analyzeSchema = analyze;
exports._compileRNG = _compileRNG;
exports._expandIncludes = _expandIncludes;
exports._isTextNodeEmpty = _isTextNodeEmpty;
exports._loadRNG = _loadRNG;
exports._lookupRNG = _lookupRNG;
exports.deserializeXMLSchema = deserializeXMLSchema;
exports.serializeXMLSchema = serializeXMLSchema;
exports.validateXML = validateXML;
